package org.apithefire.server;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apithefire.util.Objects;
import org.apithefire.util.Strings;

public abstract class Server {
	
	private final List<Connector> connectors = new ArrayList<Connector>();
	private final List<Context> contexts = new ArrayList<Context>();
	private Object serverImpl;

	private void checkNotStarted() {
		if (serverImpl != null) {
			throw new IllegalStateException("Server is started.");
		}
	}

	private void checkStarted() {
		if (serverImpl == null) {
			throw new IllegalStateException("Server is not started.");
		}
	}
	
	public void start() {
		checkNotStarted();
		
		if (connectors.isEmpty()) {
			throw new ServerException("No connector is added.");
		}
		if (contexts.isEmpty()) {
			throw new ServerException("No context is added");
		}
		for (Connector connector : connectors) {
			connector.validate();
		}
		for (Context context : contexts) {
			context.validate();
		}
		checkNoDuplicateConnectorPort();
		checkNoDuplicateContextPath();
		serverImpl = newImplementation();
		
		for (Connector connector : connectors) {
			connector.configure(serverImpl);
		}
		for (Context context : contexts) {
			context.configure(serverImpl);
		}
		start(serverImpl);
	}
	
	private void checkNoDuplicateConnectorPort() {
		HashSet<Integer> portSet = new HashSet<Integer>();
		
		for (Connector connector : connectors) {
			Integer port = connector.getPort();
			
			if (port == null) {
				// Null port should be checked prior to this
				throw new AssertionError();
			}
			if (portSet.contains(port)) {
				throw new ServerException(
						"Server contains duplicate connector port:" + port);
			} else {
				portSet.add(port);
			}
		}
	}
	
	private void checkNoDuplicateContextPath() {
		Set<String> contextPaths = new HashSet<String>();
		
		for (Context context : contexts) {
			String contextPath = Strings.emptyIfNull(context.getContextPath());
			
			if (contextPaths.contains(contextPath)) {
				throw new ServerException(
						"Server contains duplicate context path: " + contextPath);
			} else {
				contextPaths.add(contextPath);
			}
		}
	}
	
	public void join() {
		checkStarted();
		join(serverImpl);
	}
	
	public void stop() {
		checkStarted();
		stop(serverImpl);
	}

	public Connector addConnector() {
		checkNotStarted();
		
		Connector connector = Objects.nonNull(newConnector());
		
		connectors.add(connector);
		return connector;
	}

	public void removeConnector(Connector connector) {
		checkNotStarted();
		
		if (!connectors.contains(Objects.nonNull(connector))) {
			throw new IllegalArgumentException("Invalid connector.");
		}
		connectors.remove(connector);
	}

	public Context addContext() {
		checkNotStarted();
		
		Context context = Objects.nonNull(newContext());
		
		contexts.add(context);
		return context;
	}

	public void removeContext(Context context) {
		checkNotStarted();
		
		if (!connectors.contains(Objects.nonNull(context))) {
			throw new IllegalArgumentException("Invalid context.");
		}
		contexts.remove(context);
	}
	
	protected abstract Connector newConnector();
	
	protected abstract Context newContext();
	
	protected abstract Object newImplementation();
	
	protected abstract void start(Object serverImpl);
	
	protected abstract void join(Object serverImpl);
	
	protected abstract void stop(Object serverImpl);
	
}
